#!/usr/bin/perl
#==============================================================================
#
#               (c) Paul Nesfield, 2010
#
# TITLE:				Pre-Process Accu-Check Data
# VERSION:		1.5.4
# DATE:				6 April 2010
# AUTHOR:		Paul Nesfield
#
# SUMMARY
#
# A perl script to extract diabetes information from Accu-Check Smart Pix XML files.
# 
#==============================================================================

use Getopt::Std;
use Time::Local;

#==============================================================================
#
# DEFINE VARIABLES - START
#
#==============================================================================

#Define clean variables;
undef %Timestamp;
undef @Records;
undef %LocStr;

undef %MeterUnits;
undef %DeviceSerial;
undef %DeviceDates;
undef %LastDeviceDates;

undef $Events;
undef %EventNames;
undef %BolusTypes;

undef %Date;
undef %Time;
undef %Glucose;
undef %StandardTime;
  
undef %Flag;
undef %Carbohydrates;
undef %Glucose;
undef %InsulinA;
undef %InsulinB;
undef %InsulinC;
undef %EventType;
undef %GlucoseUnits;

undef %PumpBasalCBRF;
undef %PumpBasalTBRdec;
undef %PumpBasalTBRinc;
undef %PumpBasalProfile;
undef %PumpBasalRemark;

undef %PumpBolusType;
undef %PumpBolusCMD;
undef %PumpBolusAmount;
undef %PumpInsulinTotalAmount;
undef %PumpBolusRemark;

undef %PumpEventShortinfo;
undef %PumpEventDescription;

undef $LogData;



getopts('dL:P:T:');

# Parameters:
# L: string of last import datas
# P: path to the XML file or folder holding XML files
# d: if present then report the latest date of events for each meter

#==========================================
# LOCALISABLE VARIABLES - PLEASE MODIFY
#==========================================
#Retrieve the logbook data direct from the Mac OS X defaults system
undef $locdatastr;
open(LOC, $opt_L);
while (<LOC>) {
  $locdatastr .= $_."\n";
}
close(LOC);
$locdatastr =~ s/\c@//g;
$locdatastr =~ s/\r/\n/g;
$locdatastr =~ s/\f/\n/g;
$locdatastr =~ s/\cM/\n/g;
$locdatastr =~ s/.*\/\*.*\*\///g;
(@locdata) = split /"\s*;\s*\n/, $locdatastr;

foreach (@locdata) {
  $locstr=$_;
  $locstr=~ s/^\s*"//;
  ($key, $value) = split /"\s*=\s*"/, $locstr;
  $LocStr{$key}=$value if ($key);
}

%BolusTypes= (
	"Std"	=>	LocStr("Standard bolus"),
	"Scr"	=>	LocStr("Scroll bolus"),
	"Ext"	=>  LocStr("Extended bolus"),
	"Mul"	=>	LocStr("Multiwave bolus")
);

%EventNames = (
	0	=>	LocStr("No event"),
	1	=>	LocStr("Before Meal"),
	2	=>	LocStr("After Meal"),
	3	=>	LocStr("Snack"),
	4	=>	LocStr("Bed Time"),
	5	=>	LocStr("Wakeup Time"),
	6	=>	LocStr("Night Time"),
	7	=>	LocStr("(This Event ID is not used)"),
	8	=>	LocStr("Link Last Rec."),
	9	=>	LocStr("Feel Hypo."),
	10	=>	LocStr("Before Breakfast"),
	11	=>	LocStr("After Breakfast"),
	12	=>	LocStr("Before Lunch"),
	13	=>	LocStr("After Lunch"),
	14	=>	LocStr("Before Dinner"),
	15	=>	LocStr("After Dinner"),
	16	=>	LocStr("(This Event ID is not used)"),
	17	=>	LocStr("Different Food"),
	18	=>	LocStr("Over Ate"),
	19	=>	LocStr("Under Ate"),
	20	=>	LocStr("Fasting"),
	21	=>	LocStr("Skipped Meal"),
	22	=>	LocStr("Drank Alcohol"),
	23	=>	LocStr("Before Exercise"),
	24	=>	LocStr("After Exercise"),
	25	=>	LocStr("Active"),
	26	=>	LocStr("Inactive"),
	27	=>	LocStr("Vacation"),
	28	=>	LocStr("Feel High Sugar"),
	29	=>	LocStr("Stress"),
	30	=>	LocStr("Infection"),
	31	=>	LocStr("Illness"),
	32	=>	LocStr("Menses"),
	33	=>	LocStr("Invalid Test"),
	34	=>	LocStr("Others Result"),
	35	=>	LocStr("Oral Medication"),
	36	=>	LocStr("User Defined"),
	37	=>	LocStr("New Time Zone"),
	38	=>	LocStr("Hot Outdoors"),
	39	=>	LocStr("Cold Outdoors"),
	40	=>	LocStr("High Humidity"),
	41	=>	LocStr("L1 Control"),
	42	=>	LocStr("L3 Control"),
	43	=>	LocStr("L2 Control"),
	44	=>	LocStr("New Pump Tubing"),
	45	=>	LocStr("During Exercise"),
	46	=>	LocStr("(This Event ID is not used)"),
	47	=>	LocStr("C4"),
	48	=>	LocStr("C5"),
	49	=>	LocStr("C6"),
	50	=>	LocStr("C7"),
	51	=>	LocStr("C8"),
	52	=>	LocStr("C9"),
	53	=>	LocStr("(This Event ID is not used)"),
	54	=>	LocStr("D1"),
	55	=>	LocStr("D2"),
	56	=>	LocStr("D3"),
	57	=>	LocStr("D4"),
	58	=>	LocStr("D5"),
	59	=>	LocStr("D6"),
	60	=>	LocStr("D7"),
	61	=>	LocStr("D8"),
	62	=>	LocStr("D9"),
	63	=>	LocStr("D10"),
	64	=>	LocStr("D11"),
	65	=>	LocStr("D12"),
	66	=>	LocStr("D13"),
	67	=>	LocStr("D14"),
	68	=>	LocStr("No Meal Event"),
	69	=>	LocStr("Continuous Measurement"),
	70	=>	LocStr("Spot Measurement"),
	71	=>	LocStr("Hypo Warning"),
	72	=>	LocStr("Hyper Warning"),
	73	=>	LocStr("Premenstrual"),
	74	=>	LocStr("Other Meal Time"),
	75	=>	LocStr("Exercise 1"),
	76	=>	LocStr("Exercise 2"),
	"M"	=>	LocStr("Manually entered test result"),
	"M1"	=>	LocStr("Before meal"),
	"M3"	=>	LocStr("After meal"),
	"Hy"	=>	LocStr("Hypo symptomes")
);

%FlagTypes= (
	"A"	=>	LocStr("Above target range (controls only)"),
	"B"	=>	LocStr("Below target range (controls only)"),
	"D"	=>	LocStr("Used drum"),
	"E"	=>	LocStr("Expired strip"),
	"M"	=>	LocStr("Manually entered"),
	"T"	=>	LocStr("Temperature warning"),
	"*"	=>	LocStr("AST-Flag")
);

#==========================================
# NORMAL VARIABLES - DO NOT MODIFY
#==========================================

#Load the supplied last import datas for each device serial number
if ($opt_T) {
  (@LastImportDates) = split ";", $opt_T;
  foreach (@LastImportDates) {
    ($serial,$lastdate) = split ":", $_;
    $LastDeviceDates{$serial}=$lastdate;
  }
}

$PathToFile="";
$PathToFile=$opt_P if ($opt_P);

#==============================================================================
#
# DEFINE VARIABLES - END
#
#==============================================================================


#==============================================================================
#
# MAIN PROCESS - START
#
#==============================================================================

#======================
# Extract Data
#======================

#Get name of XML files in the folder
opendir(DIR, $PathToFile);
  @xmlfiles = readdir(DIR);
closedir (DIR);

#Read all the data in all the files
$Events = "";
foreach (@xmlfiles) {
  if ($_ =~ /XML$/) {
    GetData($PathToFile."/".$_);
  }
}

#Split the input data into separate events at the event delimiter {dlbxreturn}
(@Records) = split "{dlbxreturn}" , $Events;

#Process each event in the log
foreach (@Records) {

  #Prepare fresh variables
  undef %Eventdata;
  undef %MeterUnits;
  undef $Year;
  undef $Month;
  undef $Day;
  undef $Hour;
  undef $Minute;
    
  #The event data is passed as the current $_
  LoadEvent();
}


#======================
# Generate Formatted Data
#======================

#Create the formated data text
if ($opt_d) {
  CreateLatestDates();
}
else {
  CreateData();
}

print $LogData;


#==============================================================================
#
# MAIN PROCESS - END
#==============================================================================





#==============================================================================
#
# SUBROUTINES - GET, PARSE and LOAD DATA
#
#==============================================================================

#======================================================================
# Load data from the user defaults.
#======================================================================
sub GetData () {

  #Path to file is the 1st array element passed from the function call
  $myfile=@_[0];

  #Retrieve the logbook data direct from the Mac OS X defaults system
  open(DEFS, $myfile);
  @eventdata = <DEFS>;
  close(DEFS);

  #Parse each line and determine if it is a record, start assumption is 'not'
  $recordevent="NO";
  $meternum=0;
  $pumpnum=100;
  $ignorepartialbasalbolusevents=0;
  foreach (@eventdata) {
  
    #Strip off new-line and end-of-line characters
    $event=$_;
    $event=~ s/\r//;
    $event=~ s/\n//;

    #Need to capture the glucose units  and serial number per meter
    if ($event =~ /<DEVICE /) {
      $meternum++;
      ($string1, $string2) = split / BGUnit="/ , $event;
      ($glucoseunits, $string3) = split /"/ , $string2;
      $MeterUnits{$meternum}=$glucoseunits;
      ($string1, $string2) = split / SN="/ , $event;
      ($serialnum, $string3) = split /"/ , $string2;
      $serialnum=~ s/\D//g;
      $DeviceSerial{$meternum}=$serialnum;
    }
    
    #Need to capture the serial number per pump
    if ($event =~ /<IP /) {
      $pumpnum++;
      ($string1, $string2) = split / SN="/ , $event;
      ($serialnum, $string3) = split /"/ , $string2;
      $serialnum=~ s/\D//g;
      $DeviceSerial{$pumpnum}=$serialnum;
    }
    
    #Only need to parse the lines inside the 'DATA' blocks
    if ($recordevent eq "NEXT") {
      if ($ignorepartialbasalbolusevents eq 2) {
        $recordevent="YES";
      }
      else {
        $ignorepartialbasalbolusevents++;
      }
    }
    else {
      if ($event eq "<BGDATA>") {
        $recordevent="NEXT";
        $ignorepartialbasalbolusevents=2;
      }
      if ($event eq "<IPDATA>") {
        $recordevent="NEXT";
      }
      if ($event eq "</BGDATA>" || ($event eq "</IPDATA>")) {
        $recordevent="NO";
      }
  	}
  	
  	#These are data lines so process them
    if ($recordevent eq "YES") {
 
       #Work out which serial number to attach by identifying the device number
       $devicenumber=0;
       if ($event =~ /<BG /) {
        ($string1, $string2) = split / D="/ , $event;
        ($devicenumber, $string3) = split /"/ , $string2;
      }
      else {
        $devicenumber=101;
      }
 
      #Strip out unwanted characters and correct formatting, add new key/value delimitter and field delimiter
      $event=~ s/^\s*<(\w+) /\1" /;
      $event=~ s/"\s*\/>$//;
      $event=~ s/="---"/=""/g;
      $event=~ s/=" /="/g;
      $event=~ s/Tm=""/Tm="23:59"/g;
      $event=~ s/" /{dlbxtab}/g;
      $event=~ s/="/{dlbxkey}/g;
      #Add glucose units for the applicable meter device
      $event .= "{dlbxtab}Gunits{dlbxkey}".$MeterUnits{$devicenumber} if ($devicenumber >= 1 && $devicenumber <= 100);

      #Add serial number and event element delimiter {dlbxreturn}
      $Events .= "SerialNum{dlbxkey}".$DeviceSerial{$devicenumber}."{dlbxtab}".$event."{dlbxreturn}";

    }
  }

  # !!! All event data is now in the string $Events !!!

}


#=========================================
# Load extracted data into the Hash tables
#=========================================
sub LoadEvent {

  #Each imported data record is passed to this routine and loaded into the data hash tables

	my $Timestamp;
	my %Eventdata;
	
  #Split log at the {dlbxtab} delimiter marks
  (@Elements) = split "{dlbxtab}";
  
  #Load the event data into key/value hash table
  foreach (@Elements) {
    (@Field) = split "{dlbxkey}";
    $Eventdata{$Field[0]}=$Field[1];
    #Process date and time to ensure format: YYYY, MM, DD, hh, mm
    if ($Field[0] eq "Dt") {
      ($Year, $Month, $Day) = split "-", $Field[1];
    }
    if ($Field[0] eq "Tm") {
      ($Hour, $Minute) = split ":", $Field[1];
    }
  }

  $UncorrectedTimestamp=$Year.$Month.$Day.$Hour.$Minute;
  
  StandardTime();
  
  #Build a timestamp to index the Hash tables
  $Timestamp=$Year.$Month.$Day.$Hour.$Minute;
    
  #Load up the hash tables
  $Date{$Timestamp}=$Day."/".$Month."/".$Year;
  $Time{$Timestamp}=$Hour.":".$Minute;
  
  #Only process events if they are newer than the device's last import date
  $SerialNum=$Eventdata{"SerialNum"};
  if ($LastDeviceDates{$SerialNum} < $UncorrectedTimestamp) {
    $DeviceDates{$SerialNum}=$Timestamp if ($DeviceDates{$SerialNum} < $Timestamp);
  
    $Flag{$Timestamp}=$Eventdata{"Flg"} if ($Eventdata{"Flg"});
    $Carbohydrates{$Timestamp}=$Eventdata{"Carb"} if ($Eventdata{"Carb"});
    $Glucose{$Timestamp}=$Eventdata{"Val"} if (!$Eventdata{"Ctrl"} && $Eventdata{"Val"});
    $InsulinA{$Timestamp}=$Eventdata{"Ins1"} if ($Eventdata{"Ins1"});
    $InsulinB{$Timestamp}=$Eventdata{"Ins2"} if ($Eventdata{"Ins2"});
    $InsulinC{$Timestamp}=$Eventdata{"Ins3"} if ($Eventdata{"Ins3"});
    $EventType{$Timestamp}=$EventNames{scalar ($Eventdata{"Evt"})} if ($Eventdata{"Evt"});
    $EventType{$Timestamp}=$EventNames{($Eventdata{"Evt"})} if (!$EventType{$Timestamp} && ($Eventdata{"Evt"}));
    $EventType{$Timestamp}=$Eventdata{"Evt"} if (!$EventType{$Timestamp});
    $GlucoseUnits{$Timestamp}=$Eventdata{"Gunits"} if (!$Eventdata{"Ctrl"} && $Eventdata{"Val"} && $Eventdata{"Gunits"});

    if ($Eventdata{"TBRdec"}) {
      $Eventdata{"remark"} = "TBR -".$Eventdata{"TBRdec"}." ".$Eventdata{"remark"};
    }
    $PumpBasalTBRinc{$Timestamp}=$Eventdata{"TBRinc"};
    if ($Eventdata{"TBRinc"}) {
      $Eventdata{"remark"} = "TBR +".$Eventdata{"TBRinc"}." ".$Eventdata{"remark"};
    }
    $PumpBasalProfile{$Timestamp}=$Eventdata{"profile"} if ($Eventdata{"profile"});
    if ($Eventdata{"cbrf"}) {
      if ($PumpBasalRemark{$Timestamp}) {
        $PumpBasalRemark{$Timestamp}=$Eventdata{"remark"}."{lf}".$PumpBasalRemark{$Timestamp};
      }
      else {
        $PumpBasalRemark{$Timestamp}=$Eventdata{"remark"};
      }
    }

    $PumpBolusCMD{$Timestamp}=$Eventdata{"cmd"} if ($Eventdata{"cmd"});
    if ($Eventdata{"remark"} eq "Bolus+Basal Total") {
      $PumpInsulinTotalAmount{$Timestamp}=$Eventdata{"amount"};
      $PumpBolusRemark{$Timestamp}="Basal Total";
    }
    elsif ($Eventdata{"remark"} eq "Bolus Total") {
      $PumpBolusTotalAmount{$Timestamp}=$Eventdata{"amount"};
    }
    elsif ($Eventdata{"amount"}) {
      $InsulinA{$Timestamp}=$Eventdata{"amount"};
      if ($Eventdata{"remark"}) {
        $Eventdata{"remark"}=$BolusTypes{scalar ($Eventdata{"type"})}."{lf}".$Eventdata{"remark"};
      }
      else {
        $Eventdata{"remark"}=$BolusTypes{scalar ($Eventdata{"type"})};
      }
      if ($PumpBolusRemark{$Timestamp}) {
        $PumpBolusRemark{$Timestamp}=$Eventdata{"remark"}."{lf}".$PumpBolusRemark{$Timestamp};
      }
      else {
        $PumpBolusRemark{$Timestamp}=$Eventdata{"remark"};
      }
    }

    $PumpEventShortinfo{$Timestamp}=$Eventdata{"shortinfo"} if ($Eventdata{"shortinfo"});
    $PumpEventDescription{$Timestamp}=$Eventdata{"description"} if ($Eventdata{"description"});

  }
}


#=========================================
# Load extracted data into formatted text
#=========================================
sub CreateData {

  undef $LogData;
  
  #Get time ordered list of timestamps
  (@Timestamp)=sort NormalSortNum (keys %Date);
  
  $LogData .= "PFlag,Date,Time,Period,Glucose,GlucoseUnits,InsulinA,InsulinAName,Carbohydrates,InsulinB,InsulinBName,Medicine,MedicineName,KetonesTested,KetonesValue,Weight,Height,HbA1c,EventName,Notes";

  foreach (@Timestamp) {

    $AddLine="NO";
    $LogDataLine = "\n";
    
    #Flag
    $LogDataLine .= "";
    $LogDataLine .= ",";

    #Date, Time & Period
    $LogDataLine .=  $Date{$_};
    $LogDataLine .= ",";
    $LogDataLine .=  $Time{$_};
    $LogDataLine .= ",";
    $LogDataLine .= "";
    $LogDataLine .= ",";

    #Glucose & Units
    if ($Glucose{$_}) {
      $LogDataLine .= sprintf("%0.1f", $Glucose{$_});
      $AddLine="YES";
    }
    $LogDataLine .= ",";
    $LogDataLine .= $GlucoseUnits{$_};
    $LogDataLine .= ",";

    #Insulin A
    if ($InsulinA{$_}) {
      $LogDataLine .= sprintf("%0.2f", $InsulinA{$_});
	  $AddLine="YES";
	}
    $LogDataLine .= ",";
    $LogDataLine .= "";
    $LogDataLine .= ",";

    #Carbohydrates
    if ($Carbohydrates{$_}) {
      $LogDataLine .= sprintf("%0.1f", $Carbohydrates{$_});
      $AddLine="YES";
    }
    $LogDataLine .= ",";

    #Insulin B
    $InsulinB{$_} = $PumpInsulinTotalAmount{$_} - $PumpBolusTotalAmount{$_} if ($PumpInsulinTotalAmount{$_});
    if ($InsulinB{$_}) {
      $LogDataLine .= sprintf("%0.2f", $InsulinB{$_});
      $AddLine="YES";
    }
    $LogDataLine .= ",";
    $LogDataLine .= "";
    $LogDataLine .= ",";

    #Medicine
    if ($InsulinC{$_}) {
      $LogDataLine .= sprintf("%0.2f", $InsulinC{$_});
      $AddLine="YES";
    }
    $LogDataLine .= ",";
    $LogDataLine .= "";
    $LogDataLine .= ",";

    #Ketones Tested & Value
    $LogDataLine .= "";
    $LogDataLine .= ",";
    $LogDataLine .= "";
    $LogDataLine .= ",";

    #Weight
    $LogDataLine .= "";
    $LogDataLine .= ",";

    #Height
    $LogDataLine .= "";
    $LogDataLine .= ",";

    #HbA1c
    $LogDataLine .= "";
    $LogDataLine .= ",";

    #Eventname
    if ($EventType{$_}) {
      $LogDataLine .= $EventType{$_};
      $AddLine="YES";
    }
    $LogDataLine .= ",";

    #Notes
    if ($PumpBolusRemark{$_} || $PumpBasalRemark{$_} || $PumpEventDescription{$_}) {
      $LogDataLine .= $PumpBolusRemark{$_};
      $LogDataLine .= "{lf}" if ($PumpBolusRemark{$_} && $PumpBasalRemark{$_});
      $LogDataLine .= $PumpBasalRemark{$_};
      $LogDataLine .= "{lf}" if ($PumpEventDescription{$_} && ($PumpBolusRemark{$_} || $PumpBasalRemark{$_}));
      $LogDataLine .= $PumpEventDescription{$_};
      $AddLine="YES";
    }
    
    $LogData .= $LogDataLine if ($AddLine eq "YES");

  }

}

#=========================================
# Load latest event dates per device into formatted text
#=========================================
sub CreateLatestDates {

  undef $LogData;
  
  (@DeviceSN) = sort NormalSortNum (keys %DeviceDates);
  foreach (@DeviceSN) {
    $LastDeviceDates{$_}=$DeviceDates{$_} if ($DeviceDates{$_} > $LastDeviceDates{$_});
  }

  (@DeviceSN) = sort NormalSortNum (keys %LastDeviceDates);
  foreach (@DeviceSN) {
    $LogData .= $_.":".$LastDeviceDates{$_}.";";
  }

}

#=========================================
# Make time correction database for basal insulins to glucose tests
#=========================================
sub StandardTime {

  my $Time = $Year.$Month.$Day.$Hour.$Minute;
  my $myHour = $Hour;
  my $myMinute = $Minute;
  
  if ($StandardTime{$Time}) {
    ($Year,$Month,$Day,$Hour,$Minute) = split ":", $StandardTime{$Time};
  }
  else {
  	$TimeDataString=$Year.":".$Month.":".$Day.":".$Hour.":".$Minute;
  	for ($count = 1; $count <= 5; $count++) {
  	  if ($myHour < 10) {
  	    $myStrHour = "0".int $myHour;
  	  }
  	  else {
  	    $myStrHour = int $myHour;
  	  }
  	  if ($myMinute < 10){
  	    $myStrMinute = "0".int $myMinute;
  	  }
  	  else {
  	    $myStrMinute = int $myMinute;
  	  }
      $IncludedTime = $Year.$Month.$Day.$myStrHour.$myStrMinute;
      $StandardTime{$IncludedTime}=$TimeDataString;
      if ($myHour < 23 && ($myMinute < 58)) {
        $myHour++ if ($myMinute == 59);
        $myMinute++;
      }
    }
  }

}

#==============================================================================
#
# SUBROUTINES - GET, PARSE and LOAD DATA   --- END
#
#==============================================================================



#==============================================================================
#
# SUBROUTINES
#
#==============================================================================

#=======================================
# Sort function - alphabetically
#=======================================
sub LocStr {
  return $LocStr{$_[0]} if ($LocStr{$_[0]});
  return $_[0];
}

#=======================================
# Sort function - alphabetically
#=======================================
sub NormalSortNum { $a <=> $b };
